#!/usr/bin/python

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()
from routio.messages.cli import main

if __name__ == '__main__':
    main()





from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from builtins import bytes
from builtins import int
from future import standard_library
standard_library.install_aliases()
from asyncio import Future, get_event_loop

try:
    from tornado.ioloop import IOLoop
    from tornado.iostream import StreamClosedError
    import tornado.web
except ImportError:
    raise ImportError("Tornado not installed")

import routio.camera as camera

_encode_jpeg = None

try:
    from turbojpeg import TurboJPEG, TJCS_RGB

    _jpeg = TurboJPEG()

    def _turbo_encode_jpeg(image):
        return _jpeg.encode(image, quality=80, pixel_format=TJCS_RGB)

    _encode_jpeg = _turbo_encode_jpeg

except ImportError:
    try:
        import cv2
        def _cv_encode_jpeg(image):
            result, data = cv2.imencode('.jpg', image, [int(cv2.IMWRITE_JPEG_QUALITY), 80])
            if result == False:
                return None
            return bytes(data)

        _encode_jpeg = _cv_encode_jpeg

    except ImportError:
        pass

import routio

class TornadoClientObserver(routio.IOBaseObserver):

    def __init__(self, ioloop, client, disconnect_callback):
        super().__init__()
        self._ioloop = ioloop
        self._client = client
        self._canwrite = False
        self._disconnect_callback = disconnect_callback
        self._client.observe(self)

        if isinstance(ioloop, IOLoop):
            ioloop.add_handler(client.fd(), self._ioevent, IOLoop.READ | IOLoop.ERROR)


    def on_output(self, _):
        if self._canwrite:
            return

        self._ioloop.update_handler(self._client.fd(), IOLoop.READ | IOLoop.ERROR | IOLoop.WRITE)

    def _ioevent(self, _, events):
        if events & IOLoop.ERROR:
            self._client.disconnect()
            if self._disconnect_callback:
                self._self_disconnect_callback(self._client)
        else:
            if events & IOLoop.READ:
                self._client.handle_input()
            if events & IOLoop.WRITE:
                if self._client.handle_output():
                    self._ioloop.update_handler(self._client.fd(), IOLoop.READ | IOLoop.ERROR)
                    self._canwrite = False


def install_client(ioloop, client, disconnect_callback=None):
    TornadoClientObserver(ioloop, client, disconnect_callback)

def uninstall_client(ioloop, client):
    ioloop.remove_handler(client.fd())

class Image(object):
    def __init__(self, frame):
        self._raw = frame.image
        self._timestamp = frame.header.timestamp
        self._jpeg = None

    def raw(self):
        return self._raw

    def timestamp(self):
        return self._timestamp

    def jpeg(self):
        if _encode_jpeg is None:
            raise RuntimeError("No support for JPEG encoding provided")
        if self._jpeg is None:
            self._jpeg = _encode_jpeg(self._raw)
        return self._jpeg

class FutureCallback(object):

    def __init__(self):
        self._future = Future()

    def __call__(self, _, result):
        self._future.set_result(result)

    async def wait(self):
        return await self._future

class Camera(object):
    def __init__(self, client, name):
        self.name = name
        self._image_listeners = []
        self._location_listeners = []
        self._client = client
        self._parameters = None
        self._parameters_sub = camera.CameraIntrinsicsSubscriber(client, "%s.parameters" % name, self._parameters_callback)
        self._location_sub = None
        self._image_sub = None

    async def image(self):
        cb = FutureCallback()
        self.listen_images(cb)
        result = await cb.wait()
        self.unlisten_images(cb)
        return result

    async def location(self):
        cb = FutureCallback()
        self.listen_location(cb)
        result = await cb.wait()
        self.unlisten_location(cb)
        return result

    def listen_images(self, listener):
        self._image_listeners.append(listener)
        if len(self._image_listeners) == 1 and self._image_sub is None:
            self._image_sub = camera.FrameSubscriber(self._client, "%s.image" % self.name,self._frame_callback)

    def unlisten_images(self, listener):
        self._image_listeners.remove(listener)
        if len(self._image_listeners) == 0 and not self._image_sub is None:
            self._image_sub = None

    def listen_location(self, listener):
        self._location_listeners.append(listener)
        if len(self._location_listeners) == 1 and self._location_sub is None:
            self._location_sub = camera.CameraExtrinsicsSubscriber(self._client, "%s.location" % self.name, self._location_callback)

    def unlisten_location(self, listener):
        self._location_listeners.remove(listener)
        if len(self._location_listeners) == 0 and not self._location_sub is None:
            self._location_sub = None

    def _distribute_image(self, image):
        for c in self._image_listeners:
            c(self, image)
            
    def _distribute_location(self, location):
        for c in self._location_listeners:
            c(self, location)

    def _frame_callback(self, frame):
        if len(self._image_listeners) > 0:
            img = Image(frame)
            self._distribute_image(img)

    def _location_callback(self, location):
        self._distribute_location(location)

    def _parameters_callback(self, parameters):
        self._parameters = parameters

    @property
    def parameters(self):
        return self._parameters

class VideoHandler(tornado.web.RequestHandler):

    def __init__(self, application, request, camera):
        super(VideoHandler, self).__init__(application, request)
        self.camera = camera
        self.boundary = '--imageboundary--'
        self._future = None
        self._running = True

    async def get(self):
        # TODO: add random chars to boundary

        self.set_header('Content-Type', 'multipart/x-mixed-replace; boundary=' + self.boundary)
        self.flush()


        self._future = Future()
        self.camera.listen_images(self._image_cb)

        try:
            while self._running:
                image = await self._future
                self._future = Future()
                if len(self.boundary) < 1:
                    return
                
                data = image.jpeg()

                self.write(self.boundary)
                self.write("\r\n")
                self.write("Content-type: image/jpeg\r\n")
                self.write("Content-length: %d\r\n\r\n" % len(data))
                self.write(data)
                self.write("\r\n")
                await self.flush()
        except StreamClosedError:
            pass

        self.camera.unlisten_images(self._image_cb)

    def _image_cb(self, _, image):
        if not self._future.done():
            self._future.set_result(image)

    def on_finish(self):
        self._running = False

    def on_connection_close(self):
        self._running = False

    def check_etag_header(self):
        return False

class ImageHandler(tornado.web.RequestHandler):
    def __init__(self, application, request, camera):
        super(ImageHandler, self).__init__(application, request)
        self.camera = camera

    def set_default_headers(self):
        self.set_header('Content-Type', 'image/jpeg')
        self.set_header('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0')

    async def get(self):
        image = await self.camera.image()
        self.set_header('X-Timestamp', image.timestamp().isoformat())
        self.write(image.jpeg())
        self.finish()

    def check_etag_header(self):
        return False


__all__ = ["Camera", "ImageHandler", "VideoHandler", "install_client", "uninstall_client"]


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from future import standard_library
standard_library.install_aliases()
import routio
import numpy

try:
    from routio import _wrapper
except ImportError as ie:
    raise ImportError("Echo binary library not found", ie)

class CameraIntrinsics(object):
    def __init__(self):
        self.width = None
        self.height = None
        self.intrinsics = None
        self.distortion = None
        self.name = 'Unknown'

    @staticmethod
    def read(reader):
        obj = CameraIntrinsics()
        obj.width = reader.readInt()
        obj.height = reader.readInt()
        obj.intrinsics = _wrapper.readTensor(reader)
        obj.distortion = _wrapper.readTensor(reader)
        return obj

    @staticmethod
    def write(writer, obj):
        writer.writeInt(obj.width)
        writer.writeInt(obj.height)
        _wrapper.writeTensor(writer, obj.intrinsics)
        _wrapper.writeTensor(writer, obj.distortion)

class CameraExtrinsics(object):

    def __init__(self, header = routio.Header(), rotation=numpy.eye(3), translation=numpy.empty((1, 3))):
        self.header = header
        self.rotation = rotation
        self.translation = translation

    @staticmethod
    def read(reader):
        obj = CameraExtrinsics()
        obj.header = routio.readType(routio.Header, reader)
        obj.rotation = _wrapper.readTensor(reader)
        obj.translation = _wrapper.readTensor(reader)
        return obj

    @staticmethod
    def write(writer, obj):
        routio.writeType(routio.Header, writer, obj.header)
        _wrapper.writeTensor(writer, obj.rotation)
        _wrapper.writeTensor(writer, obj.translation)

routio.registerType(numpy.array, _wrapper.readTensor, _wrapper.writeTensor)
routio.registerType(CameraIntrinsics, CameraIntrinsics.read, CameraIntrinsics.write)
routio.registerType(CameraExtrinsics, CameraExtrinsics.read, CameraExtrinsics.write)

class Frame(object):

    def __init__(self, header = routio.Header(), image = numpy.array(())):
        self.header = header
        self.image = image

    @staticmethod
    def read(reader):
        obj = Frame()
        obj.header = routio.readType(routio.Header, reader)
        obj.image = _wrapper.readTensor(reader)
        return obj

    @staticmethod
    def write(writer, obj):
        routio.writeType(routio.Header, writer, obj.header)
        _wrapper.writeTensor(writer, obj.image)

routio.registerType(Frame, Frame.read, Frame.write)

class CameraIntrinsicsSubscriber(routio.Subscriber):

    def __init__(self, client, alias, callback):
        def _read(message):
            reader = routio.MessageReader(message)
            return CameraIntrinsics.read(reader)

        super().__init__(client, alias, "camera intrinsics", lambda x: callback(_read(x)))


class CameraIntrinsicsPublisher(routio.Publisher):

    def __init__(self, client, alias):
        super().__init__(client, alias, "camera intrinsics")

    def send(self, obj):
        writer = routio.MessageWriter()
        CameraIntrinsics.write(writer, obj)
        super().send(writer)

class CameraExtrinsicsSubscriber(routio.Subscriber):

    def __init__(self, client, alias, callback):
        def _read(message):
            reader = routio.MessageReader(message)
            return CameraExtrinsics.read(reader)

        super().__init__(client, alias, "camera extrinsics", lambda x: callback(_read(x)))


class CameraIntrinsicsPublisher(routio.Publisher):

    def __init__(self, client, alias):
        super().__init__(client, alias, "camera extrinsics")

    def send(self, obj):
        writer = routio.MessageWriter()
        CameraExtrinsics.write(writer, obj)
        super().send(writer)

class FrameSubscriber(routio.Subscriber):

    def __init__(self, client, alias, callback):
        def _read(message):
            reader = routio.MessageReader(message)
            return Frame.read(reader)

        super().__init__(client, alias, "camera frame", lambda x: callback(_read(x)))

class FramePublisher(routio.Publisher):

    def __init__(self, client, alias, queue=1):
        super().__init__(client, alias, "camera frame", queue)

    def send(self, obj):
        writer = routio.MessageWriter()
        Frame.write(writer, obj)
        super().send(writer)


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import open
from future import standard_library
standard_library.install_aliases()
import sys
import os
import argparse
import logging

import jinja2

from routio.messages import MessagesRegistry, parseFile, formatConstant, DescriptionError, set_default_language
from routio.messages.library import library_path
from routio.messages.templates import templates_path

logger = logging.getLogger("echomsg")

def render_language(language, extension, registry, outdir, basename):
    loader = jinja2.FileSystemLoader(templates_path())
    env = jinja2.Environment(loader=loader)
    env.filters["constant"] = lambda x: formatConstant(x, language=language)
    template = env.get_template('{}.tpl'.format(language))

    set_default_language(language)
    
    context = {}
    context['basename'] = basename
    context['namespace'] = registry.namespace
    context['registry'] = registry

    if not outdir is None:
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        with open(os.path.join(outdir, "{}.{}".format(context['basename'], extension)), "w") as out:
            out.write(template.render(context))

def main():

    logger.addHandler(logging.StreamHandler())

    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('files', metavar='file', type=str, nargs='+',
                        help='Message files to process')

    parser.add_argument("--debug", "-d", default=False, help="Backup backend", required=False, action='store_true')

    parser.add_argument('-p', metavar='paths', type=str, action='append', default=[],
                        help='Message files to process', dest='paths')

    parser.add_argument('--python-outdir', metavar='dir', type=str,
                        help='Output directory for Python files', dest='outdir_python')

    parser.add_argument('--cpp-outdir', metavar='dir', type=str,
                        help='Output directory for C++ files', dest='outdir_cpp')

    args = parser.parse_args()

    args.paths.insert(0, library_path())
    args.paths.append(".")

    logger.setLevel(logging.INFO)

    if args.debug:
        logger.setLevel(logging.DEBUG)

    for filename in args.files:
        logger.debug("Processing file %s", filename)
        registry = MessagesRegistry()
        registry.namespace = ''
        try:
            parseFile(filename, registry, args.paths)
        except DescriptionError as e:
            logger.error("Processing error: %s", e)
            sys.exit(1)

        logger.debug("Generating C++ wrapper")
        render_language("cpp", "h", registry, args.outdir_cpp, os.path.splitext(os.path.basename(filename))[0])
        logger.debug("Generating Python wrapper")
        render_language("python", "py", registry, args.outdir_python, os.path.splitext(os.path.basename(filename))[0])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import map
from builtins import super
from future import standard_library
standard_library.install_aliases()
from pyparsing import *

class DescriptionError(Exception):
    
    def __init__(self, file, line, column, message):
        super().__init__()
        self._file = file
        self._line = line
        self._column = column
        self._message = message

    def __str__(self):
        return "{} (line: {}, col: {}): {}".format(self._file, self._line, self._column, self._message)

    @property
    def file(self):
        return self._file

    @property
    def line(self):
        return self._line

    @property
    def column(self):
        return self._column

    @property
    def message(self):
        return self._message

def description_syntax():

    def make_keyword(kwd_str, kwd_value):
        return Keyword(kwd_str).setParseAction(replaceWith(kwd_value))

    TRUE = make_keyword("true", True)
    FALSE = make_keyword("false", False)
    NULL = make_keyword("null", None)

    LBRACK, RBRACK, LBRACE, RBRACE, COLON, SEMICOLON, EQUALS, POINT, LANGLE, RANGLE = map(
        Suppress, "[]{}:;=.()")

    Exponent = CaselessLiteral('E')

    MessageFile = dblQuotedString().setParseAction(removeQuotes)
    StringLiteral = dblQuotedString().setParseAction(removeQuotes)

    PlusMinus = Literal('+') | Literal('-')
    Number = Word(nums)
    IntegerValue = Combine( Optional(PlusMinus) + Number )
    FloatValue = Combine( IntegerValue +
                       Optional( POINT + Optional(Number) ) +
                       Optional( Exponent + IntegerValue )
                     )
    BooleanValue = Or( [TRUE | FALSE])

    LanguageName = Word(alphanums)
    FieldName = Word(alphanums + "_")
    FieldType = Word(alphanums)
    PropertyName = Word(alphanums + "_")
    EnumerateConst = Word(nums)
    EnumerateValue = Group(Word(alphanums + "_").setResultsName("name"))
    EnumerateName = Word(alphanums + "_")
    StructureName = Word(alphanums + "_")
    ArrayLength = Word(nums)
    
    Value = Group(Or([FloatValue.setResultsName("numeric") | StringLiteral.setResultsName("string") | BooleanValue.setResultsName("bool")]))
    PositionalProperty = Value.setResultsName("value")
    KeywordProperty = Group(PropertyName.setResultsName("name") + EQUALS + Value.setResultsName("value"))
    PropertyList = Group(LANGLE +
        (Group(KeywordProperty + ZeroOrMore(Suppress(COLON) + KeywordProperty).setResultsName("kwargs")) ^
        (Group(PositionalProperty + ZeroOrMore(Suppress(COLON) + PositionalProperty)).setResultsName("args") + ZeroOrMore(Suppress(COLON) + KeywordProperty).setResultsName("kwargs"))) + RANGLE)

    ExternalLanguage = Group(Literal("language") + LanguageName.setResultsName("language") 
            + StringLiteral.setResultsName("container") 
            + Optional(Literal("from") + OneOrMore(StringLiteral).setResultsName("sources"))
            + Optional(Literal("default") + StringLiteral.setResultsName("default"))
            + Optional(Literal("read") + StringLiteral.setResultsName("read") 
                + Literal("write") + StringLiteral.setResultsName("write"))
            ) + SEMICOLON

    ExternalLanguageList = Group(LANGLE + ZeroOrMore(ExternalLanguage) + RANGLE)

    Field = Group(FieldType.setResultsName("type") + Optional(Group(LBRACK +
        Optional(ArrayLength).setResultsName("length") + RBRACK).setResultsName("array")) +
        FieldName.setResultsName("name") + Optional(PropertyList.setResultsName("properties")) + Optional(EQUALS + Value.setResultsName("default")) + SEMICOLON)
    FieldList = Group(
        LBRACE + ZeroOrMore(Field) + RBRACE).setResultsName('fields')

    Enumerate = Group(Literal("enumerate") + EnumerateName.setResultsName("name") + LBRACE +
                      Group(delimitedList(EnumerateValue)).setResultsName('values') + RBRACE)

    Include = Group(
        Literal("include") + MessageFile.setResultsName('name') + Optional(PropertyList.setResultsName("properties")))  + SEMICOLON

    Import = Group(
        Literal("import") + MessageFile.setResultsName('name')) + SEMICOLON

    External = Group(
        Literal("external") + StructureName.setResultsName('name') + ExternalLanguageList.setResultsName("languages")) + SEMICOLON

    Structure = Group(
        Literal("structure") + StructureName.setResultsName('name') + FieldList)
    Message = Group(
        Literal("message") + StructureName.setResultsName('name') + FieldList)

    Namespace = Group(
        Literal("namespace") + Word(alphanums + ".").setResultsName('name') + SEMICOLON)
    Messages = Optional(
        Namespace) + ZeroOrMore(Or([Enumerate | Include | Import | External | Structure | Message]))
    Messages.ignore(pythonStyleComment)

    return Messages

def parse(text):

    syntax = description_syntax()

    try:
        return syntax.parseString(text, True)
    except ParseException as e:
        raise DescriptionError("<input>", e.lineno, e.col, e.msg)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from future import standard_library
standard_library.install_aliases()
from routio import _wrapper

class ArraySubscriber(_wrapper.Subscriber):

    def __init__(self, client, alias, callback):
        def _read(message):
            reader = _wrapper.MessageReader(message)
            return _wrapper.readArray(reader)

        super().__init__(client, alias, "array", lambda x: callback(_read(x)))

class ArrayPublisher(_wrapper.Publisher):

    def __init__(self, client, alias):
        super().__init__(client, alias, "array")

    def send(self, obj):
        writer = _wrapper.MessageWriter()
        _wrapper.writeArray(writer, obj)
        super().send(writer)

class TensorSubscriber(_wrapper.Subscriber):

    def __init__(self, client, alias, callback):
        def _read(message):
            reader = _wrapper.MessageReader(message)
            return _wrapper.readTensor(reader)

        super().__init__(client, alias, "tensor", lambda x: callback(_read(x)))

class TensorPublisher(_wrapper.Publisher):

    def __init__(self, client, alias):
        super().__init__(client, alias, "tensor")

    def send(self, obj):
        writer = _wrapper.MessageWriter()
        _wrapper.writeTensor(writer, obj)
        super().send(writer)


from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from builtins import int
from builtins import super
from builtins import range
from builtins import str
from future import standard_library
from future.utils import bytes_to_native_str

standard_library.install_aliases()
try:
    import routio.pyroutio as _wrapper
except ImportError as ie:
    raise ImportError("Routio binary library not found", ie)

IOLoop = _wrapper.IOLoop
IOBase = _wrapper.IOBase
IOBaseObserver = _wrapper.IOBaseObserver
Client = _wrapper.Client
Publisher = _wrapper.Publisher
Subscriber = _wrapper.Subscriber
MessageReader = _wrapper.MessageReader
MessageWriter = _wrapper.MessageWriter

double = type(bytes_to_native_str(b'double'), (), {})
char = type(bytes_to_native_str(b'char'), (), {})
long = type(bytes_to_native_str(b'long'), (), {})

_type_registry = {}

def router(address="/tmp/routio.sock", verbose=False):
    _wrapper.router(address, verbose)

def registerType(cls, read, write, convert=None):
    if cls.__name__ in _type_registry:
        raise Exception("Type %s already registered" % cls.__name__)
    _type_registry[cls.__name__] = {"read" : read, "write" : write, "convert" : convert}

def readType(cls, reader):
    return _type_registry[cls.__name__]["read"](reader)

def writeType(cls, writer, obj):
    if not isinstance(obj, cls):
        if _type_registry[cls.__name__]["convert"] is None:
            raise Exception("Object type is not correct %s != %s" % (type(obj).__name__, cls.__name__))
        try:
            obj = _type_registry[cls.__name__]["convert"](obj)
        except Exception as e:
            raise Exception("Unable to convert %s to %s: %s" % (type(obj).__name__, cls.__name__, e))
    _type_registry[cls.__name__]["write"](writer, obj)

def convertInt(obj):
    return int(obj)

def convertFloat(obj):
    return float(obj)

def convertStr(obj):
    return str(obj)

def convertBool(obj):
    return bool(obj)

registerType(int, lambda x: x.readInt(), lambda x, o: x.writeInt(o), convertInt)
registerType(long, lambda x: x.readLong(), lambda x, o: x.writeLong(o), convertInt)
registerType(float, lambda x: x.readFloat(), lambda x, o: x.writeFloat(o), convertFloat)
registerType(str, lambda x: x.readString(), lambda x, o: x.writeString(o), convertStr)
registerType(double, lambda x: x.readDouble(), lambda x, o: x.writeDouble(o), convertFloat)
registerType(char, lambda x: x.readChar(), lambda x, o: x.writeChar(o))
registerType(bool, lambda x: x.readBool(), lambda x, o: x.writeBool(o), convertBool)

import datetime

registerType(datetime.datetime, _wrapper.readTimestamp, _wrapper.writeTimestamp)

def readList(cls, reader):
    objects = []
    n = reader.readInt()
    read = _type_registry[cls.__name__]["read"]
    for i in range(n):
        objects.append(read(reader))
    return objects

def writeList(cls, writer, objects):
    writer.writeInt(len(objects))
    write = _type_registry[cls.__name__]["write"]
    for o in objects:
        write(writer, o)

class Dictionary(dict):

    def __init__(self, reader=None):
        super().__init__()
        if reader:
            n = reader.readInt()
            for _ in range(0, n):
                key = reader.readString()
                value = reader.readString()
                self[key] = value

    @staticmethod
    def read(reader):
        obj = Dictionary()
        n = reader.readInt()
        for _ in range(0, n):
            key = reader.readString()
            value = reader.readString()
            obj[key] = value
        return obj

    @staticmethod
    def write(writer, obj):
        writer.writeInt(len(obj))
        for k, v in obj.items():
            writer.writeString(k)
            writer.writeString(str(v))

    def pack(self):
        writer = _wrapper.MessageWriter(4 + sum([ len(k) + len(v) + 8 for k, v in self.items()]))
        self.write(writer)
        return writer

    def __setitem__(self, key, item):
        self.__dict__[key] = item

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key]

    def get(self, key, default = None):
        return self.__dict__.get(key, default)

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        return self.__dict__.copy()

    def has_key(self, k):
        return self.__dict__.has_key(k)

    def pop(self, key, default = None):
        return self.__dict__.pop(key, default)

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

registerType(Dictionary, Dictionary.read, Dictionary.write)

class Header(object):
    def __init__(self, source = "", timestamp = None):
        self.source = source
        if timestamp == None:
            self.timestamp = datetime.datetime.now()
        else:
            self.timestamp = timestamp

    @staticmethod
    def read(reader):
        return Header(readType(str, reader), readType(datetime.datetime, reader))

    @staticmethod
    def write(writer, obj):
        writeType(str, writer, obj.source)
        writeType(datetime.datetime, writer, obj.timestamp)

registerType(Header, Header.read, Header.write)

class DictionarySubscriber(_wrapper.Subscriber):

    def __init__(self, client, alias, callback):
        def _read(message):
            reader = _wrapper.MessageReader(message)
            return Dictionary.read(reader)

        super().__init__(client, alias, "dictionary", lambda x: callback(_read(x)))

class DictionaryPublisher(_wrapper.Publisher):

    def __init__(self, client, alias):
        super().__init__(client, alias, "dictionary")

    def send(self, obj):
        writer = _wrapper.MessageWriter()
        Dictionary.write(writer, obj)
        super().send(writer)

class SubscriptionWatcher(_wrapper.Watcher):
    def __init__(self, client, alias, callback=None, watch=True):
        super().__init__(client, alias)
        #self._callback = callback
        #self._subscribers = 0
        #if watch:
        #    self.watch()

    def on_event(self, message):
        print("event", message)
        #mtype = message.get("type", "")

        #if mtype == "subscribe" or mtype == "unsubscribe" or mtype == "summary":
        #    try:
        #        self._subscribers = int(message.get("subscribers", 0))
        #    except ValueError:
        #        self._subscribers = 0

            #if self._callback is not None:
            #    self._callback(self._subscribers)

    @property
    def subscribers(self):
        return self._subscribers



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future import standard_library
standard_library.install_aliases()
import os
import argparse
import sys

def main():
    from routio import router

    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbose", action="store_true", default=False)
    parser.add_argument("bind", nargs='?', default=os.environ.get("ROUTIO_SOCKET", "/tmp/routio.sock"))

    try:
        args = parser.parse_args()
    except argparse.ArgumentError:
        parser.print_help()
        sys.exit(-1)

    try:
        router(args.bind, args.verbose)
    except KeyboardInterrupt:
        pass

if __name__ == "__main__":
    main()
    


import sys
import os
import glob
import setuptools

from setuptools import setup

# Ensure optimized compilation flags
if 'CXXFLAGS' not in os.environ:
    os.environ['CXXFLAGS'] = '-O3 -DNDEBUG -s'
if 'CFLAGS' not in os.environ:
    os.environ['CFLAGS'] = '-O3 -DNDEBUG -s'

__version__ = "1.0.0"

from distutils.core import setup, Extension
from distutils.command.build_ext import build_ext

class get_pybind_include(object):
    def __str__(self):
        import pybind11
        return pybind11.get_include()

class get_numpy_include(object):
    def __str__(self):
        import numpy
        return numpy.get_include()

root = os.path.abspath(os.path.dirname(__file__))
platform = os.getenv("ROUTIO_PYTHON_PLATFORM", sys.platform)

if platform.startswith('linux'):
    library_suffix = '.so'
elif platform in ['darwin']:
    library_suffix = '.dylib'
elif platform.startswith('win'):
    library_suffix = '.dll'

class build_ext_ctypes(build_ext):

    def build_extension(self, ext):
        self._ctypes = isinstance(ext, CTypes)

        def _realize(c):
            from typing import Callable
            if isinstance(c, Callable):
                return c()
            else:
                return c

        ct = self.compiler.compiler_type
        opts = ["-std=c++20"]

        ext.include_dirs = [_realize(x) for x in ext.include_dirs]
        ext.extra_compile_args = opts

        return super().build_extension(ext)

    def get_export_symbols(self, ext):
        if self._ctypes:
            return ext.export_symbols
        return super().get_export_symbols(ext)

    def get_ext_filename(self, ext_name):
        if self._ctypes:
            return ext_name + library_suffix
        return super().get_ext_filename(ext_name)

class CTypes(Extension): 
    pass

varargs = dict()
varargs["package_data"] = {}
varargs["cmdclass"] = {}

try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

    class bdist_wheel(_bdist_wheel):

        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            # Mark us as not a pure python package
            self.root_is_pure = False

        def get_tag(self):
            python, abi, plat = _bdist_wheel.get_tag(self)
            # We don't contain any python source
            python, abi = 'py2.py3', 'none'
            return python, abi, plat

        def run(self):
            # Force rebuild of extensions with optimization flags
            self.run_command('build_ext')
            
            # Temporarily exclude C++ source files from wheels
            original_exclude = varargs.get("exclude_package_data", {})
            varargs["exclude_package_data"] = {'routio.source': ['*.cpp']}
            varargs["exclude_packages"] = ['routio.source']

            try:
                _bdist_wheel.run(self)
            finally:
                # Restore original exclude_package_data
                if original_exclude:
                    varargs["exclude_package_data"] = original_exclude
                else:
                    varargs.pop("exclude_package_data", None)

    varargs["cmdclass"] = {'bdist_wheel': bdist_wheel, 'build_ext': build_ext_ctypes}
    varargs["setup_requires"] = ['wheel']

except ImportError:
    pass

# Detect which command is being run
if os.path.isfile(os.path.join("routio", "pyroutio" + library_suffix)):
    # Binary exists - include it for wheel builds, exclude for source distributions
    varargs["package_data"]["routio"] = ["pyroutio" + library_suffix]
    # Exclude the routio.source package (contains C++ source files)

elif os.path.isfile(os.path.join("routio", "source", "loop.cpp")):
    # Source files exist - for sdist, compile from source; for wheels, build the extension
    sources = glob.glob("routio/source/*.cpp")
    args = dict(sources=sources, include_dirs=[os.path.join("routio", "source"), get_numpy_include(), get_pybind_include()], define_macros=[("ROUTIO_EXPORTS", "1")])
    args["language"] = "c++"
    varargs["ext_modules"] = [CTypes("routio.pyroutio", **args)]
    varargs["cmdclass"] = {'build_ext': build_ext_ctypes}
    varargs["setup_requires"] = ["pybind11>=2.5.0", "numpy>=1.16"]
    # Include headers and source for source distributions (sdist)
    varargs["package_data"]['routio.source'] = ['*.h', '*.cpp']
    varargs["package_data"]['routio.source.routio'] = ['*.h']

varargs["package_data"]['routio.messages.library'] = ['*.msg']
varargs["package_data"]['routio.messages.templates'] = ['*.tpl']

setup(
    name='routio',
    version=__version__,
    author='Luka Cehovin Zajc',
    author_email='luka.cehovin@protonmail.com',
    url='https://github.com/lukacu/routio',
    project_urls={
        'Documentation': 'https://github.com/lukacu/routio',
        'Source Code': 'https://github.com/lukacu/routio',
        'Bug Tracker': 'https://github.com/lukacu/routio/issues',
    },
    description='Python package for routio IPC library',
    long_description='A simple and efficient interprocess communication (IPC) library for Linux written in C++11.',
    long_description_content_type='text/plain',
    keywords=['ipc', 'interprocess', 'communication', 'messaging', 'rpc'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: C++',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Distributed Computing',
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires='>=3.10',
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'routio-daemon = routio.__main__:main',
            'routio-router = routio.__main__:main',
            'routio-generate = routio.messages.cli:main',
        ],
        'ignition': [
            'routio_remap = routio.ignition:Mapping',
        ],
    },
    **varargs
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import super
from builtins import open
from builtins import str
from builtins import int
from builtins import zip
from builtins import range
from future import standard_library
standard_library.install_aliases()
import os
import six
import hashlib
import logging

from collections import OrderedDict

from routio.messages.syntax import parse, DescriptionError

LANGUAGES = ["cpp", "python"]

default_language = "cpp"

logger = logging.getLogger("echomsg")

def set_default_language(language):
    if language in LANGUAGES:
        global default_language
        default_language = language

def remove_duplicates(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


class Type(object):

    def __init__(self, name, hash):
        self._name = name
        self._hash = hash

    def get_name(self):
        return self._name

    def get_container(self, language=None):
        return self._name

    def get_default(self, language=None):
        return None

    def get_reader(self, language=None):
        return None

    def get_writer(self, language=None):
        return None

    def get_hash(self):
        return self._hash

class ExternalType(Type):     
    def __init__(self, name, container, default = None, reader = None, writer = None):
        super(ExternalType, self).__init__(name, name)
        self._default = default
        self._container = container
        self._reader = reader
        self._writer = writer

    def get_container(self, language=None):
        language = language if language else default_language
        if isinstance(self._container, dict):
            return self._container.get(language, None)
        return self._container

    def get_default(self, language=None):
        language = language if language else default_language
        if isinstance(self._default, dict):
            return self._default.get(language, None)
        return self._default

    def get_reader(self, language=None):
        language = language if language else default_language
        if isinstance(self._reader, dict):
            return self._reader.get(language, None)
        return self._reader

    def get_writer(self, language=None):
        language = language if language else default_language
        if isinstance(self._writer, dict):
            return self._writer.get(language, None)
        return self._writer

class Source(object):

    def __init__(self, code):
        self.code = code

    def __str__(self):
        return self.code


def formatConstant(value, language="cpp"):
    if isinstance(value, Source):
        return str(value)
    if value is None and language == "cpp":
        return ""
    if value is None and language == "python":
        return "None"
    elif isinstance(value, bool):
        if language == "cpp":
            return "true" if value else "false"
        if language == "python":
            return "True" if value else "False"
    if isinstance(value, six.integer_types):
        return str(value)
    if isinstance(value, float):
        return "%.f" % (value)
    elif isinstance(value, six.string_types):
        return "\"%s\"" % value
    return str(value)

class MessagesRegistry(object):

    def __init__(self):
        self.enums = OrderedDict()
        self.types = OrderedDict()
        self.add_type(ExternalType("short", {"python" : "int", "cpp": "int16_t"}, 0))
        self.add_type(ExternalType("int", {"python" : "int", "cpp": "int32_t"}, 0))
        self.add_type(ExternalType("long", {"python" : "routio.long", "cpp": "int64_t"}, 0))
        self.add_type(ExternalType("float", {"python" : "float", "cpp": "float"}, 0.0))
        self.add_type(ExternalType("double", {"python" : "routio.double", "cpp": "double"}, 0.0))
        self.add_type(ExternalType("bool", {"python" : "bool", "cpp": "bool"}, False))
        self.add_type(ExternalType("char", {"python" : "routio.char", "cpp": "char"}, '\0'))
        self.add_type(ExternalType("string", {"python" : "str", "cpp": "std::string"}, ""))

        self.add_type(ExternalType("timestamp", {
            "python" : "datetime.datetime",
            "cpp": "std::chrono::system_clock::time_point"
        }))

        self.add_type(ExternalType("header", {
            "python" : "routio.Header",
            "cpp": "routio::Header"
        }, default={
            "python" : Source("routio.Header()"),
            "cpp": Source("routio::Header()")
        }))

        self.add_type(ExternalType("array", {
            "python" : "numpy.ndarray",
            "cpp": "routio::Array"
        }, default={
            "python" : Source("numpy.zeros((0,))"),
            "cpp": Source("routio::Array()")
        }))

        self.add_type(ExternalType("tensor", {
            "python" : "numpy.ndarray",
            "cpp": "routio::Array"
        }, default={
            "python" : Source("numpy.zeros((0,))"),
            "cpp": Source("routio::Tensor()")
        }))

        self.structs = OrderedDict()
        self.messages = []
        self.namespace = None
        self.files = []
        self.sources = {l : [] for l in LANGUAGES}
        self.sources["cpp"].append("vector")
        self.sources["cpp"].append("chrono")
        self.sources["cpp"].append("routio/datatypes.h")
        self.sources["cpp"].append("routio/array.h")
        self.sources["python"].append("routio")
        self.sources["python"].append("datetime")
        self.sources["python"].append("numpy")
        
    def get_sources(self, language = None):
        language = language if language else default_language
        return self.sources[language]

    def add_enum(self, name, values):
        
        typehash = hashlib.md5()
        for v in values:
            typehash.update(v.encode('utf-8'))
        self.add_type(Type(name, typehash.hexdigest()))
        self.enums[name] = values

    def add_struct(self, name, fields):
        typehash = hashlib.md5()
        for k, v in fields.items():
            t = v['type']
            if not t in self.types:
                raise RuntimeError('Unknown type: ' + t)
            typehash.update(self.types[t].get_hash().encode('utf-8'))

        self.add_type(Type(name, typehash.hexdigest()))
        self.structs[name] = fields

    def add_message(self, name, fields):
        self.add_struct(name, fields)
        self.messages.append(name)

    def add_type(self, type):
        if type.get_name() in self.types:
            raise RuntimeError('Name already taken: ' + type.get_name())
        self.types[type.get_name()] = type

def processValue(value):
    if "numeric" in value:
        try:
            return int(value["numeric"])
        except ValueError:
            return float(value["numeric"])
    elif "bool" in value:
        return bool(value["bool"])
    else:
        return str(value["string"])

def processFields(fields):
    result = OrderedDict()
    for field in fields:
        name = field["name"]
        result[name] = {"type": field["type"], "default" : None}
        result[name]["array"] = "array" in field
        if result[name]["array"]:
            result[name]["array"] = True
            if "length" in field["array"]:
                result[name]["length"] = int(field["array"]["length"])
#        if "properties" in field:
#           result[name]["properties"]
        if "default" in field:
           result[name]["default"] = processValue(field["default"])
    return result


def parseFile(msgfile, registry, searchpath=[], include=True):

    if os.path.isabs(msgfile):
        if not os.path.exists(msgfile) or not os.path.isfile(msgfile):
            raise IOError("File '%s' does not exist", msgfile)
    else:
        base = msgfile
        msgfile = None
        for p in searchpath:
            candidate = os.path.join(p, base)
            if os.path.exists(candidate) and os.path.isfile(candidate):
                msgfile = candidate
                break

    if msgfile in registry.files:
        logger.info("%s already processed, ignoring.", msgfile)
        return

    registry.files.append(msgfile)

    if not msgfile:
        raise IOError("File '%s' does not exist in search path" % base)

    try:
        with open(msgfile, "r") as handle:
            messages = parse(handle.read())
    except DescriptionError as e:
        raise DescriptionError(msgfile, e.line, e.column, e.message)

    try:

        for e in messages:
            name = e[0]
            if name == 'namespace':
                registry.namespace = e["name"]
            if name == 'import':
                parseFile(e["name"], registry, searchpath, False)
            if name == 'include':
                properties = {a["name"] : a["value"] for a in e.get("properties", [])}
                parseFile(e["name"], registry, searchpath, True)
            if name == 'external':
                containers = {l: e["name"] for l in LANGUAGES}
                defaults = {}
                readers = {}
                writers = {}
                declared = []
                for l in e.get("languages", []):
                    if not l["language"] in LANGUAGES:
                        raise RuntimeError("Unknown language {}".format(l["language"]))
                    if l["language"] in declared:
                        raise RuntimeError("Duplicate declaration {}".format(l["language"]))
                    declared.append(l["language"])
                    containers[l["language"]] = l["container"] # Override container name
                    if l.get("default", None):
                        defaults[l["language"]] = Source(l["default"])
                    if l.get("sources", None):
                        registry.sources[l["language"]].extend(l["sources"])
                    if l.get("read", None):
                        readers[l["language"]] = Source(l["read"])
                        writers[l["language"]] = Source(l["write"])

                registry.add_type(ExternalType(e["name"], containers, defaults, readers, writers))
            if name == 'enumerate':
                values = {v["name"]: k for v, k in zip(e["values"], range(len(e["values"])))}
                registry.add_enum(e["name"], values)
            if name == 'structure':
                fields = processFields(e["fields"])
                registry.add_struct(e["name"], fields)
            if name == 'message':
                fields = processFields(e["fields"])
                registry.add_message(e["name"], fields)

        registry.sources = {k: remove_duplicates(l) for k, l in registry.sources.items()}

    except RuntimeError as ex:
        logger.exception(ex)
        raise DescriptionError(msgfile, 0, 0, str(ex))





from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from future import standard_library
standard_library.install_aliases()

try:
    # Python 3.10 changes
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

try:
    from ignition.plugin import Plugin
except ImportError:
    raise ImportError("Ignition package is required to use this module")

class Mapping(Plugin):

    def __init__(self):
        super().__init__()

    def on_program_init(self, program):
        mapping = program.auxiliary.get("remap", {})

        if not isinstance(mapping, Mapping):
            return
        maplist = []
        for k, v in mapping.items():
            maplist.append("%s=%s" % (k, v))

        program.environment["ROUTIO_MAP"] = ";".join(maplist)

